// This source is subject to the Microsoft Public License (Ms-PL).
// Please see http://go.microsoft.com/fwlink/?LinkID=131993 for details.
// All other rights reserved.

using System;
using System.Reflection;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Navigation;
using TimePunch.Metro.Wpf.Controls.Picker;
using TimePunch.Metro.Wpf.EventAggregation;
using TimePunch.Metro.Wpf.Events;
using TimePunch.Metro.Wpf.Frames;
using TimePunch.Metro.Wpf.Helper;
using TimePunch.Metro.Wpf.ViewModel;

namespace TimePunch.Metro.Wpf.Controller
{
    /// <summary>
    /// Base class for module specific controllers
    /// </summary>
    public abstract class BaseController :
        IHandleMessage<GoBackNavigationRequest>,
        IHandleMessage<CloseApplicationCommand>,
        IHandleMessage<ListPickerFullModeRequest>,
        IHandleMessage<DatePickerFullModeRequest>,
        IHandleMessage<TimePickerFullModeRequest>,
        IHandleMessage<TimeSpanPickerFullModeRequest>,
        IHandleMessage<WindowStateApplicationCommand>,
        IHandleMessage<ForceBindingUpdateEvent>,
        IHandleMessage<ChangeAnimationModeRequest>, 
        INavigationController
    {
        private NavigationMode navigationMode;

        #regionConstructors

        /// <summary>
        /// Initializes a new instance of the BaseController class.
        /// </summary>
        protected BaseController()
        {
            EventAggregator = Kernel.Instance.EventAggregator;
            EventAggregator.Subscribe(this);
        }

        /// <summary>
        /// This method can do some initializations
        /// </summary>
        public virtual void Init()
        {
        }

        /// <summary>
        /// Handles a message of a specific type.
        /// </summary>
        /// <param name="message">the message to handle</param>
        public void Handle(ForceBindingUpdateEvent message)
        {
            var focusObj = FocusManager.GetFocusedElement(Application.Current.MainWindow);
            if (focusObj == null || !(focusObj is TextBox)) 
                return;
            
            var binding = (focusObj as TextBox).GetBindingExpression(TextBox.TextProperty);

            if (binding != null) 
                binding.UpdateSource();
        }

        /// <summary>
        /// Releases unmanaged resources and performs other cleanup operations before the
        /// <see cref="BaseController"/> is reclaimed by garbage collection.
        /// </summary>
        ~BaseController()
        {
            EventAggregator.Unsubscribe(this);
        }

        #endregionConstructors

        #regionMethods

        /// <summary>
        /// Attaches the controller to the root frame
        /// </summary>
        /// <param name="contentFrame">The Content Frame</param>
        public virtual void SetRootFrame(Frame contentFrame)
        {
            if (contentFrame != null)
            {
                // wire events
                contentFrame.Navigated += OnNavigated;
                contentFrame.Navigating += OnNavigating;
            }

            ContentFrame = contentFrame;
        }

        /// <summary>
        /// Gets or sets the content frame
        /// </summary>
        protected Frame ContentFrame { get; set; }

        /// <summary>
        /// Called when the NavigationService navigates to a new page inside the application
        /// </summary>
        /// <param name="sender">The sending object</param>
        /// <param name="e">Navigating Event Args</param>
        protected virtual void OnNavigated(object sender, NavigationEventArgs e)
        {
            var page = e.Content as Page;

            // keep track of the current page
            CurrentPage = page;
            if (page == null) return;

            // Send NavigationCompletedEvent, if it's a normal navigation request
            object context = page.DataContext;
            if (context != null)
            {
                // Send Navigated Request
                Type requestType = typeof (NavigationCompletedEvent<>).MakeGenericType(new[] {context.GetType()});

                MethodInfo publishMessage = EventAggregator.GetType().GetMethod("PublishMessage");
                publishMessage = publishMessage.GetGenericMethodDefinition().MakeGenericMethod(new[] {requestType});
                publishMessage.Invoke(EventAggregator, new[] { Activator.CreateInstance(requestType, navigationMode) });
            }

            // Initializes the Page if the DataContext is derived from the ViewModelBase
            var viewModelBase = page.DataContext as ViewModelBase;
            if (viewModelBase != null)
                viewModelBase.InitializePage(e.ExtraData);

            EventAggregator.PublishMessage(new NavigatedEvent());
        }

        /// <summary>
        /// On Navigation Handler
        /// </summary>
        /// <param name="sender">The sending object</param>
        /// <param name="e">Navigating Cancel Event Args</param>
        protected virtual void OnNavigating(object sender, NavigatingCancelEventArgs e)
        {
            navigationMode = e.NavigationMode;
            EventAggregator.PublishMessage(new NavigationEvent(e));
        }

        /// <summary>
        /// Navigates to page.
        /// </summary>
        /// <param name="navigateToPage">The navigate to page.</param>
        public virtual void NavigateToPage(string navigateToPage)
        {
            if (CurrentPage == null)
                Application.Current.Dispatcher.BeginInvoke((ThreadStart)(() =>
                    ContentFrame.Navigate(new Uri(navigateToPage, UriKind.Relative))));
            else
            {
                if (CurrentPage.Dispatcher.CheckAccess())
                {
                    if (CurrentPage.NavigationService == null)
                        return;

                    PasswordBoxAssistant.IsNavigating = true;
                    try
                    {
                        CurrentPage.NavigationService.Navigate(new Uri(navigateToPage, UriKind.Relative));
                    }
                    finally
                    {
                        PasswordBoxAssistant.IsNavigating = false;
                    }
                }
                else
                    CurrentPage.Dispatcher.BeginInvoke((ThreadStart) (() => NavigateToPage(navigateToPage)));
            }
        }

        /// <summary>
        /// Navigates to a page and add additional data
        /// </summary>
        /// <param name="navigateToPage">The navigate to page.</param>
        /// <param name="message">The message thas will be send to the page</param>
        public virtual void NavigateToPage(string navigateToPage, object message)
        {
            if (CurrentPage == null)
                Application.Current.Dispatcher.BeginInvoke((ThreadStart)(() => 
                    ContentFrame.Navigate(new Uri(navigateToPage, UriKind.Relative), message)));
            else
            {
                if (CurrentPage.Dispatcher.CheckAccess())
                {
                    if (CurrentPage.NavigationService == null)
                        return;

                    PasswordBoxAssistant.IsNavigating = true;
                    try
                    {
                        CurrentPage.NavigationService.Navigate(new Uri(navigateToPage, UriKind.Relative), message);
                    }
                    finally
                    {
                        PasswordBoxAssistant.IsNavigating = false;    
                    }
                }
                else
                    CurrentPage.Dispatcher.BeginInvoke((ThreadStart) (() => NavigateToPage(navigateToPage, message)));
            }
        }

        /// <summary>
        /// Determines whether this instance [can go back].
        /// </summary>
        /// <returns>
        /// 	<c>true</c> if this instance [can go back]; otherwise, <c>false</c>.
        /// </returns>
        public virtual bool CanGoBack
        {
            get { return CurrentPage != null && CurrentPage.NavigationService != null && CurrentPage.NavigationService.CanGoBack; }
        }

        #endregion

        #regionProperties

        /// <summary>
        /// Gets or sets the controller that used for monitor the navigational messages
        /// </summary>
        protected IEventAggregator EventAggregator { get; set; }

        /// <summary>
        /// Gets or sets the current page.
        /// </summary>
        /// <value>The current page.</value>
        public Page CurrentPage { get; private set; }

        #endregionProperties

        #region IHandleMessage<GoBackNavigationRequest> Members

        /// <summary>
        /// Handles the Navigation Request
        /// </summary>
        /// <param name="message">The Navigation goback request</param>
        public void Handle(GoBackNavigationRequest message)
        {
            // Now Go Back
            if (CurrentPage.Dispatcher.CheckAccess())
            {
                // When going back, delete an potential error and clear the state
                var baseModel = CurrentPage.DataContext as ViewModelBase;
                if (baseModel != null && baseModel.IsDefective)
                    baseModel.Error = string.Empty;

                if (CurrentPage.NavigationService == null)
                    return;

                if (CurrentPage.NavigationService.CanGoBack)
                    CurrentPage.NavigationService.GoBack();
            }
            else
                CurrentPage.Dispatcher.BeginInvoke((ThreadStart)(() => Handle(message)));
        }

        #endregion

        #region IHandleMessage<CloseApplicationCommand>

        /// <summary>
        /// Handles a message of a specific type.
        /// </summary>
        /// <param name="message">the message to handle</param>
        public virtual void Handle(CloseApplicationCommand message)
        {
            ForceClosing = message.ForceClosing;
            Application.Current.MainWindow.Close();
        }

        /// <summary>
        /// Gets a value indicating whether the closing of the current window shall be forced.
        /// </summary>
        public virtual bool ForceClosing { get; private set; }

        #endregion

        #region IHandleMessage<WindowStateApplicationCommand>

        /// <summary>
        /// Handles a message of a specific type.
        /// </summary>
        /// <param name="message">the message to handle</param>
        public virtual void Handle(WindowStateApplicationCommand message)
        {
            Application.Current.MainWindow.WindowState = message.WindowState;
        }

        #endregion

        #region IHandleMessage<ListPickerFullModeRequest>

        /// <summary>
        /// Handles a message of a specific type.
        /// </summary>
        /// <param name="message">the message to handle</param>
        public virtual void Handle(ListPickerFullModeRequest message)
        {
            NavigateToPage("/TpMetroWpfLibrary;component/Controls/Picker/ListPickerFullMode.xaml", message);
        }

        #endregion

        #region IHandleMessage<DatePickerFullModeRequest>

        /// <summary>
        /// Handles a message of a specific type.
        /// </summary>
        /// <param name="message">the message to handle</param>
        public virtual void Handle(DatePickerFullModeRequest message)
        {
            NavigateToPage("/TpMetroWpfLibrary;component/Controls/Picker/DatePickerFullMode.xaml", message);
        }

        #endregion

        #region IHandleMessage<TimePickerFullModeRequest>

        /// <summary>
        /// Handles a message of a specific type.
        /// </summary>
        /// <param name="message">the message to handle</param>
        public virtual void Handle(TimePickerFullModeRequest message)
        {
            NavigateToPage("/TpMetroWpfLibrary;component/Controls/Picker/TimePickerFullMode.xaml", message);
        }

        #endregion

        #region IHandleMessage<TimePickerFullModeRequest>

        /// <summary>
        /// Handles a message of a specific type.
        /// </summary>
        /// <param name="message">the message to handle</param>
        public virtual void Handle(TimeSpanPickerFullModeRequest message)
        {
            NavigateToPage("/TpMetroWpfLibrary;component/Controls/Picker/TimeSpanPickerFullMode.xaml", message);
        }

        #endregion

        #region IHandleMessage<ChangeAnimationModeRequest>

        /// <summary>
        /// This method will change the animation mode
        /// </summary>
        /// <param name="message"></param>
        public virtual void Handle(ChangeAnimationModeRequest message)
        {
            var animationFrame = ContentFrame as AnimationFrame;
            if (animationFrame == null)
                return;

            // Switch the Animation Mode
            AnimationMode oldMode = animationFrame.AnimationMode;
            animationFrame.AnimationMode = message.AnimationMode;
            message.AnimationMode = oldMode;
        }

        #endregion
    }
}